const which = require('which');
const os = require('os');
const archiver = require('archiver');
const StreamZip = require('node-stream-zip');
const path = require('path');
const electron = require('electron');
const fs = require('fs-extra');
const hidefile = require('hidefile');
const windowStateKeeper = require('electron-window-state');
const tcpPortUsed = require('tcp-port-used');
const ipcMain = electron.ipcMain;
const isDev = require('electron-is-dev');
const {autoUpdater} = require('electron-updater')
// Module to control application life.
const app = electron.app;
// Module to create native browser window.
const BrowserWindow = electron.BrowserWindow;
const dialog = electron.dialog;

var gotLock = app.requestSingleInstanceLock();

if (!gotLock) {
	app.quit();
	return;
}

app.on('second-instance', function (commandLine, workingDirectory) {
	// Someone tried to run a second instance, we should create a new window
	fs.writeFileSync(os.homedir() + '/.postage/postage-SIGHUP', '\n', 'utf8');
});

var strResourcesPath = path.normalize(process.resourcesPath);
if (strResourcesPath.indexOf('envelope/postage_electron/node_modules/electron/dist/resources') > -1) {
	strResourcesPath = strResourcesPath.replace('envelope/postage_electron/node_modules/electron/dist/resources', 'envelope/postage_electron');
} else if (strResourcesPath.indexOf('envelope/postage_electron/node_modules/electron/dist/Electron.app/Contents/Resources') > -1) {
	strResourcesPath = strResourcesPath.replace('envelope/postage_electron/node_modules/electron/dist/Electron.app/Contents/Resources', 'envelope/postage_electron');
} else if (strResourcesPath.indexOf('envelope\\postage_electron\\node_modules\\electron\\dist\\resources') > -1) {
	strResourcesPath = strResourcesPath.replace('envelope\\postage_electron\\node_modules\\electron\\dist\\resources', 'envelope\\postage_electron');
}


// Keep a global reference of the window object, if you don't, the window will
// be closed automatically when the JavaScript object is garbage collected.
let mainWindows = [];
let globalUpdateWindow;
let mainWindowState = null;
let bolPostageIsReady = false;

try {
	fs.statSync(os.homedir() + '/.postage/');
	fs.statSync(os.homedir() + '/.postage/postage.conf');
	fs.statSync(os.homedir() + '/.postage/postage-connections.conf');
} catch (e) {
	fs.mkdirsSync(os.homedir() + '/.postage/');
	hidefile.hideSync(os.homedir() + '/.postage/');

	fs.writeFileSync(os.homedir() + '/.postage/postage-SIGHUP', '\n', 'utf8');

	console.log('copying config');
	fs.writeFileSync(os.homedir() + '/.postage/postage.conf', fs.readFileSync(strResourcesPath + '/app/postage/config/postage.conf', 'utf8'), 'utf8');
	fs.writeFileSync(os.homedir() + '/.postage/postage-connections.conf', fs.readFileSync(strResourcesPath + '/app/postage/config/postage-connections.conf', 'utf8'), 'utf8');
}

fs.writeFileSync(os.homedir() + '/.postage/postage-SIGHUP', '\n', 'utf8');
var lastTime = new Date().getTime();
fs.watch(os.homedir() + '/.postage/postage-SIGHUP', function (eventType) {
	if (eventType === 'change') {
		var currTime = new Date().getTime();
		if ((currTime - lastTime) > 1000) {
			openWindow();
		}
		lastTime = currTime;
	}
});

if (process.platform == 'win32') {
	try {
		fs.statSync(process.env.APPDATA + '\\postgresql\\pgpass.conf');
	} catch (e) {
		fs.mkdirsSync(process.env.APPDATA + '\\postgresql');
		fs.writeFileSync(process.env.APPDATA + '\\postgresql\\pgpass.conf', '\n', 'utf8');
	}
} else {
	try {
		fs.statSync(os.homedir() + '/.pgpass');
	} catch (e) {
		fs.writeFileSync(os.homedir() + '/.pgpass', '\n', 'utf8');
		fs.chmodSync(os.homedir() + '/.pgpass', 0o600);
	}
}

function spawnPostage(bolRefresh) {
	console.log(strResourcesPath, '/postage/envelope');
	proc = child_process.spawn(
		path.normalize(strResourcesPath + '/app/postage/envelope' + (process.platform == 'win32' ? '.exe' : '')),
		[
			'-c', path.normalize(os.homedir() + '/.postage/postage.conf'),
			'-d', path.normalize(os.homedir() + '/.postage/postage-connections.conf'),
			'-r', strResourcesPath + path.normalize('/app/postage/web_root'),
			'-x', 'on',
			'-p', int_postage_port,
			'-m', 'postage',
			'-f', '*',
			'-l', 'info',
			'-k', path.normalize(os.homedir() + (process.platform == 'win32' ? '/AppData/Roaming/Postage/sql' : '/.postage/sql')),
			'-3', '0',
			'-0', 'true'
		].concat((process.platform == 'win32' ? ['-o', 'stderr'] : [])), {
			detached: true
		}
	);
	console.log([
		'-c', path.normalize(os.homedir() + '/.postage/postage.conf'),
		'-d', path.normalize(os.homedir() + '/.postage/postage-connections.conf'),
		'-r', strResourcesPath + path.normalize('/app/postage/web_root'),
		'-x', 'on',
		'-p', int_postage_port,
		'-m', 'postage',
		'-f', '*',
		'-l', 'info',
		'-k', path.normalize(os.homedir() + (process.platform == 'win32' ? '/AppData/Roaming/Postage/sql' : '/.postage/sql')),
		'-3', '0',
		'-0', 'true'
	]);

	proc.stdout.on('data', function (data) {
		console.log('got data:\n' + data);
		if (data.toString().indexOf('in your web browser') > -1) {
			fs.writeFileSync(path.normalize(os.homedir() + '/.postage/postage_port'), int_postage_port.toString());
			if (mainWindows.length === 0 && !globalUpdateWindow) {
				createUpdateWindow();
			}
			bolPostageIsReady = true;
			if (bolRefresh) {
				mainWindows.forEach(function (curWindow) {
					curWindow.webContents.executeJavaScript('window.location.reload();');
				});
			}
		}
	});
	proc.stderr.on('data', function (data) {
		console.log('got data:\n' + data);
	});
	proc.on('close', function (code) {
		console.log(proc.pid + ' closed with code ' + code);
	});
}

function handleRedirect(e, url) {
	if (url.indexOf('http://127.0.0.1:' + int_postage_port) === -1) {
		e.preventDefault()
		electron.shell.openExternal(url);
	}
}

var int_postage_port = null;

function pickNewPort() {
	int_postage_port = parseInt(Math.random().toString().substring(2), 10) % (65535 - 1024) + 1024;
	tcpPortUsed.check(int_postage_port, '127.0.0.1').then(function this_callback(taken) {
		if (!taken) {
			spawnPostage();

			var localStoragePath = path.normalize(app.getPath('userData') + '/Local Storage');
			if (fs.existsSync(localStoragePath)) {
				fs.readdir(localStoragePath, function(err, items) {
					var i, len;
					if (err) {
						electron.dialog.showErrorBox('Error', err.message);
						app.quit();
					}

					for (i = 0, len = items.length; i < len; i++) {
						if (/^http_127\.0\.0\.1_([0-9]+)\.localstorage/gi.test(items[i])) {
							console.log(items[i]);
							var oldPath = path.normalize(localStoragePath + '/' + items[i]);
							var newPath = path.normalize(localStoragePath + '/' + items[i].replace(/([0-9]+)\.localstorage/gi, function (match) {
								return int_postage_port.toString() + match.substring(match.indexOf('.'));
							}));
							fs.renameSync(oldPath, newPath);
							console.log(newPath);
						}
					}
				});
			}
		} else {
			int_postage_port = parseInt(Math.random().toString().substring(2), 10) % (65535 - 1024) + 1024;
			tcpPortUsed.check(int_postage_port, '127.0.0.1').then(this_callback, function (err) {
				electron.dialog.showErrorBox('Error', err.message);
				app.quit();
			});
		}
	}, function (err) {
	    throw err;
	});
}

try {
	int_postage_port = parseInt(fs.readFileSync(path.normalize(os.homedir() + '/.postage/postage_port')), 10);
	tcpPortUsed.check(int_postage_port, '127.0.0.1').then(function (taken) {
		if (!taken) {
			spawnPostage();
		} else {
			pickNewPort();
		}
	}, function (err) {
	    throw err;
	});
} catch (e) {
	pickNewPort();
}

const child_process = require('child_process');
var proc = null;
var pgProc = null;
var strDataDirectory = null;
var pgPort;
var strInitDBPath;
var strPGCtlPath;
var strPGDumpAllPath;
var strPGDumpPath;
var strPGRestorePath;
var strPsqlPath;
var intPGProgress = 0;
var dumpProc;
var restoreProc;
if (process.platform === 'win32') {
	strInitDBPath = strResourcesPath + path.normalize('/app/postgresql/bin/initdb.exe');
	strPGCtlPath = strResourcesPath + path.normalize('/app/postgresql/bin/pg_ctl.exe');
	strPGDumpAllPath = strResourcesPath + path.normalize('/app/postgresql/bin/pg_dumpall.exe');
	strPGDumpPath = strResourcesPath + path.normalize('/app/postgresql/bin/pg_dump.exe');
	strPGRestorePath = strResourcesPath + path.normalize('/app/postgresql/bin/pg_restore.exe');
	strPsqlPath = strResourcesPath + path.normalize('/app/postgresql/bin/psql.exe');
} else if (process.platform === 'linux') {
	strInitDBPath = strResourcesPath + path.normalize('/app/postgresql/bin/initdb');
	strPGCtlPath = strResourcesPath + path.normalize('/app/postgresql/bin/pg_ctl');
	strPGDumpAllPath = strResourcesPath + path.normalize('/app/postgresql/bin/pg_dumpall');
	strPGDumpPath = strResourcesPath + path.normalize('/app/postgresql/bin/pg_dump');
	strPGRestorePath = strResourcesPath + path.normalize('/app/postgresql/bin/pg_restore');
	strPsqlPath = strResourcesPath + path.normalize('/app/postgresql/bin/psql');
} else if (process.platform === 'darwin') {
	strInitDBPath = strResourcesPath + path.normalize('/app/postgresql/bin/initdb');
	strPGCtlPath = strResourcesPath + path.normalize('/app/postgresql/bin/pg_ctl');
	strPGDumpAllPath = strResourcesPath + path.normalize('/app/postgresql/bin/pg_dumpall');
	strPGDumpPath = strResourcesPath + path.normalize('/app/postgresql/bin/pg_dump');
	strPGRestorePath = strResourcesPath + path.normalize('/app/postgresql/bin/pg_restore');
	strPsqlPath = strResourcesPath + path.normalize('/app/postgresql/bin/psql');
}

function shutDownPostgres(callback) {
	pgProc = child_process.spawn(strPGCtlPath, [
		'-D', strDataDirectory
		, 'stop'
	]);
	console.log('pgProc.spawnargs', pgProc.spawnargs);
	pgProc.stdout.on('data', function (data) {
		console.log('pg_ctl stdout', data.toString());
	});
	pgProc.stderr.on('data', function (data) {
		console.log('pg_ctl stderr', data.toString());
	});
	pgProc.on('close', function (code) {
		console.log('pg_ctl close', code);
		fs.rmSync(strDataDirectory, { recursive: true, force: true });
		pgProc = null;
		if (callback) {
			callback();
		}
	});
}

ipcMain.on('postage', function (event, arg) {
	console.log(arg);
	if (arg === 'restart') {
		proc.kill();
		spawnPostage(true);
	} else if (arg === 'init_cluster') {
		strDataDirectory = path.normalize(os.homedir() + '/.postage/temp_data' + (+new Date()));
		console.log('strDataDirectory', strDataDirectory);

		intPGProgress = 0;
		pgProc = child_process.spawn(strInitDBPath, [
			'-D', strDataDirectory
			, '-E', 'UTF8'
			, '-U', 'postgres'
			, '-d' // more output for better progress reporting
		]);
		console.log('pgProc.spawnargs', pgProc.spawnargs);
		pgProc.stdout.on('data', function (data) {
			console.log('initdb stdout', data.toString());
			intPGProgress += data.length;
			event.reply('postgres_progress', intPGProgress);
		});
		pgProc.stderr.on('data', function (data) {
			console.log('initdb stderr', data.toString());
			intPGProgress += data.length;
			event.reply('postgres_progress', intPGProgress);
		});
		pgProc.on('close', function (code) {
			console.log('initdb close', code);
			pgPort = parseInt(Math.random().toString().substring(2), 10) % (65535 - 1024) + 1024;
			tcpPortUsed.check(pgPort, '127.0.0.1').then(function this_callback(taken) {
				console.log(pgPort, taken);
				if (taken) {
					pgPort = parseInt(Math.random().toString().substring(2), 10) % (65535 - 1024) + 1024;
					tcpPortUsed.check(int_postage_port, '127.0.0.1').then(this_callback, function (err) {
						electron.dialog.showErrorBox('Error', err.message);
						app.quit();
					});
				} else {
					pgProc = child_process.spawn(strPGCtlPath, [
						'-D', strDataDirectory
						, '-o', '-p ' + pgPort + ' -c unix_socket_directories=' + path.normalize(os.homedir() + '/.postage/')
						, 'start'
					]);
					intPGProgress = 0;
					console.log('pgProc.spawnargs', pgProc.spawnargs);
					pgProc.stdout.on('data', function (data) {
						console.log('pg_ctl stdout', data.toString());
						intPGProgress += data.length;
						event.reply('postgres_progress2', intPGProgress);
						if (data.toString().indexOf('done') > -1) {
							// event.returnValue = pgPort;
							event.reply('postgres', path.normalize(os.homedir() + '/.postage/') + '&' + pgPort);
						}
					});
					pgProc.stderr.on('data', function (data) {
						console.log('pg_ctl stderr', data.toString());
						intPGProgress += data.length;
						event.reply('postgres_progress2', intPGProgress);
					});
					pgProc.on('close', function (code) {
						console.log('pg_ctl close', code);
					});
				}
			}, function (err) {
				throw err;
			});
		});

	} else if (arg.indexOf('shutdown_postgres') === 0) {
		shutDownPostgres();

	} else if (arg.indexOf('pg_restore_password') === 0) {
		var strPassword = decodeURIComponent(arg.split('&')[1]);
		restoreProc.stdin.write(strPassword + '\n');
	} else if (arg.indexOf('pg_restore') === 0) {
		var arrArg = arg.split('&');
		var jsnArg = JSON.parse(arrArg[1]);
		var strPath = dialog.showOpenDialogSync(null, {
			title: 'Restore from pg_dump Archive',
			filters: [
				{
					name: 'postage_dump Archive',
					extensions: ['postage_dump']
				},
				{
					name: 'SQL file',
					extensions: ['sql']
				},
				{
					name: 'pg_dump Archive',
					extensions: ['dump']
				}
			],
			properties: ['openFile']
		});
		if (strPath) {
			strPath = strPath[0]; // showOpenDialogSync returns an array
			if (jsnArg.host.indexOf('\\')) {
				console.log(jsnArg.host);
				jsnArg.host = path.normalize(
					jsnArg.host
						.replace(/^'/g, '')
						.replace(/'$/g, '')
						.replace(/\\'/g, '\'')
				);
				console.log(jsnArg.host);
			}
			var restoreSQLFunction = function (strPath, callback) {
				restoreProc = child_process.spawn(strPsqlPath, [
					'-h', jsnArg.host
					, '-p', jsnArg.port
					, '-U', jsnArg.user
					, '-d', jsnArg.dbname
					, '-f', strPath
				]);
				console.log('strPGRestorePath', strPGRestorePath);
				console.log('restoreProc.spawnargs', restoreProc.spawnargs);
				restoreProc.stdout.on('data', function (data) {
					console.log('pg_restore stdout', data.toString());
					event.reply('pg_restore', data.toString());
				});
				restoreProc.stderr.on('data', function (data) {
					console.log('pg_restore stderr', data.toString());
					event.reply('pg_restore', data.toString());
				});
				restoreProc.on('close', function (code) {
					event.reply('pg_restore', '[psql exited with code ' + code + ']\n');
					callback();
				});
			};
			var restoreDumpFunction = function (strPath, callback) {
				restoreProc = child_process.spawn(strPGRestorePath, [
					'-h', jsnArg.host
					, '-p', jsnArg.port
					, '-U', jsnArg.user
					, '-d', jsnArg.dbname
					, '--format=custom'
					, '-v'
					, strPath
				]);
				console.log('strPGRestorePath', strPGRestorePath);
				console.log('restoreProc.spawnargs', restoreProc.spawnargs);
				restoreProc.stdout.on('data', function (data) {
					console.log('pg_restore stdout', data.toString());
					event.reply('pg_restore', data.toString());
				});
				restoreProc.stderr.on('data', function (data) {
					console.log('pg_restore stderr', data.toString());
					event.reply('pg_restore', data.toString());
				});
				restoreProc.on('close', function (code) {
					event.reply('pg_restore', '[pg_restore exited with code ' + code + ']\n');
					callback();
				});
			};
			var restorePostageDumpFunction = function () {
				event.reply('pg_restore', '[extracting pg_dumpall.sql]\n');
				const zip = new StreamZip.async({ file: strPath });
				var strSQLPath = strPath.replace(/\.postage_dump$/, '.sql');
				var strDumpPath = strPath.replace(/\.postage_dump$/, '.dump');
				zip.extract('pg_dumpall.sql', strSQLPath).then(function () {
					restoreSQLFunction(strSQLPath, function () {
						fs.unlinkSync(strSQLPath);
						event.reply('pg_restore', '[extracting pg_dump.dump]\n');
						zip.extract('pg_dump.dump', strDumpPath).then(function () {
							restoreDumpFunction(strDumpPath, function () {
								fs.unlinkSync(strDumpPath);
								zip.close();
								event.reply('pg_restore', '[restore complete]\n');
							});
						});
					});
				});
			};
			
			if (/\.sql$/.test(strPath)) {
				restoreSQLFunction(strPath, function () {
					event.reply('pg_restore', '[restore complete]\n');
				});
			} else if (/\.dump$/.test(strPath)) {
				restoreDumpFunction(strPath, function () {
					event.reply('pg_restore', '[restore complete]\n');
				});
			} else if (/\.postage_dump$/.test(strPath)) {
				restorePostageDumpFunction();
			}
		} else {
			event.reply('pg_restore', 'cancelled');
		}

	} else if (arg.indexOf('pg_dump_password') === 0) {
		var strPassword = decodeURIComponent(arg.split('&')[1]);
		dumpProc.stdin.write(strPassword + '\n');
	} else if (arg.indexOf('pg_dump') === 0) {
		var arrArg = arg.split('&');
		var jsnArg = JSON.parse(arrArg[1]);
		var arrFilter;
		if (jsnArg.data === 'Globals, Schema, and Data' || jsnArg.data === 'Globals and Schema') {
			arrFilter = [
				{
					name: 'postage_dump Archive',
					extensions: ['postage_dump']
				}
			];
		} else if (jsnArg.data === 'Globals Only') {
			arrFilter = [
				{
					name: 'SQL file',
					extensions: ['sql']
				}
			];
		} else {
			arrFilter = [
				{
					name: 'pg_dump Archive',
					extensions: ['dump']
				}
			];
		}
		var strPath = dialog.showSaveDialogSync(null, {
			title: 'Save pg_dump Archive',
			filters: arrFilter
		});
		if (!strPath) {
			event.reply('pg_dump', 'cancelled');
			return;
		}
		if (jsnArg.host.indexOf('\\')) {
			console.log(jsnArg.host);
			jsnArg.host = path.normalize(
				jsnArg.host
					.replace(/^'/g, '')
					.replace(/'$/g, '')
					.replace(/\\'/g, '\'')
			);
			console.log(jsnArg.host);
		}
		var pgDumpAllFunction = function (callback) {
			var arrDumpArg = [
				'-h', jsnArg.host
				, '-p', jsnArg.port
				, '-U', jsnArg.user
				, '--globals-only'
				, '--verbose'
				, '-f' , strPath.replace(/\.(postage_dump|dump)$/, '.sql')
			];
			dumpProc = child_process.spawn(
				strPGDumpAllPath
				, arrDumpArg
				, {
					env: {
						// force pg_dump to skip windows console shenanigans
						OSTYPE: 'msys'
					}
				}
			);
			event.reply('pg_dump', 'pg_dumpall ' + arrDumpArg.join(' ') + '\n');
			console.log('strPGDumpAllPath', strPGDumpAllPath);
			console.log('dumpProc.spawnargs', dumpProc.spawnargs);
			dumpProc.stdout.on('data', function (data) {
				console.log('pg_dump stdout', data.toString());
				event.reply('pg_dump', data.toString());
			});
			dumpProc.stderr.on('data', function (data) {
				console.log('pg_dump stderr', data.toString());
				event.reply('pg_dump', data.toString());
			});
			dumpProc.on('close', function (code) {
				event.reply('pg_dump', '[pg_dumpall exited with code ' + code + ']\n');
				callback();
			});
		};
		var pgDumpFunction = function () {
			var arrDumpArg = [
				'-F', 'custom'
				, '-h', jsnArg.host
				, '-p', jsnArg.port
				, '-U', jsnArg.user
				, '-d', jsnArg.dbname
				, '--verbose'
				, '-f' , strPath.replace(/\.postage_dump$/, '.dump')
			];
			if (jsnArg.data.indexOf('Data') === -1) {
				arrDumpArg.push('--schema-only');
			}
			if (jsnArg.data.indexOf('Schema') === -1) {
				arrDumpArg.push('--data-only');
			}
			dumpProc = child_process.spawn(
				strPGDumpPath
				, arrDumpArg
				, {
					env: {
						// force pg_dump to skip windows console shenanigans
						OSTYPE: 'msys'
					}
				}
			);
			event.reply('pg_dump', 'pg_dump ' + arrDumpArg.join(' ') + '\n');
			console.log('strPGDumpPath', strPGDumpPath);
			console.log('dumpProc.spawnargs', dumpProc.spawnargs);
			dumpProc.stdout.on('data', function (data) {
				console.log('pg_dump stdout', data.toString());
				event.reply('pg_dump', data.toString());
			});
			dumpProc.stderr.on('data', function (data) {
				console.log('pg_dump stderr', data.toString());
				event.reply('pg_dump', data.toString());
			});
			dumpProc.on('close', function (code) {
				event.reply('pg_dump', '[pg_dump exited with code ' + code + ']\n');
				if (/\.postage_dump$/.test(strPath)) {
					event.reply('pg_dump', '[archiving to .postage_dump]\n');
					const output = fs.createWriteStream(strPath);
					const archive = archiver('zip', {
						zlib: { level: 9 } // Sets the compression level.
					});
					var strSQLPath = strPath.replace(/\.postage_dump$/, '.sql');
					var strDumpPath = strPath.replace(/\.postage_dump$/, '.dump');
					output.on('close', function() {
						event.reply('pg_dump', '[archive complete]\n');
						fs.unlinkSync(strSQLPath);
						fs.unlinkSync(strDumpPath);
						electron.shell.showItemInFolder(strPath);
					});
					archive.on('error', function(err) {
						throw err;
					});
					archive.pipe(output);
					archive.append(fs.createReadStream(strSQLPath), { name: 'pg_dumpall.sql' });
					archive.append(fs.createReadStream(strDumpPath), { name: 'pg_dump.dump' });
					archive.finalize();
				} else {
					electron.shell.showItemInFolder(strPath);
				}
			});
		};
		if (jsnArg.data === 'Globals Only') {
			pgDumpAllFunction(function () {
				electron.shell.showItemInFolder(strPath.replace(/\.dump$/, '.sql'));
			});
		} else if (jsnArg.data.indexOf('Globals') > -1) {
			pgDumpAllFunction(pgDumpFunction);
		} else {
			pgDumpFunction();
		}
	}
});

function beforeWindowClose(event) {
	if (pgProc) {
		let choice = dialog.showMessageBoxSync(null, {
			type: 'question',
			buttons: ['Stay', 'Close'],
			title: 'Closing Postage will destroy your cluster',
			message: 'If you want to save your data, you can use Tools > Dump Database',
			defaultId: 0,
			cancelId: 0
		});

		if (choice === 0) {
			event.preventDefault();
			console.log('staying');
		} else if (choice === 1) {
			console.log('closing without dumping');
			shutDownPostgres(function () {
				app.quit();
			});
		}
	}
}

app.on('before-quit', function (event) {
	if (pgProc) {
		event.preventDefault();
	}
});

app.on('quit', function () {
	proc.kill();
});

function closeTabFunc() {
	BrowserWindow.getFocusedWindow().webContents.executeJavaScript('closeCurrentTab()');
}

function setMenu() {
	const Menu = electron.Menu;
	const template = [
		{
			label: 'File',
			submenu: [
				{
					label: 'New Window',
					accelerator: 'CmdOrCtrl+N',
					click: openWindow
				},
				{
					role: 'quit'
				}
			]
		}, {
			label: 'Edit',
			submenu: [
				{
					role: 'undo'
				}, {
					role: 'redo'
				}, {
					type: 'separator'
				}, {
					role: 'cut'
				}, {
					role: 'copy'
				}, {
					role: 'paste'
				}, {
					role: 'delete'
				}, {
					role: 'selectall'
				}
			]
		}, {
			label: 'View',
			submenu: [
				{
					label: 'Reload',
					accelerator: 'CmdOrCtrl+R',
					role: 'forcereload'
				}, {
					label: 'Toggle Developer Tools',
					accelerator: process.platform === 'darwin' ? 'Alt+Command+I' : 'Ctrl+Shift+I',
					click: function (item, focusedWindow) {
						if (focusedWindow) {
							focusedWindow.webContents.toggleDevTools();
						}
					}
				}, {
					type: 'separator'
				}, {
					role: 'resetzoom'
				}, {
					role: 'zoomin'
				}, {
					role: 'zoomout'
				}, {
					type: 'separator'
				}, {
					role: 'togglefullscreen'
				}
			]
		}, {
			role: 'window',
			submenu: [
				{
					role: 'minimize'
				}, {
					label: 'Close Tab',
					accelerator: 'CmdOrCtrl+W',
					click: closeTabFunc
				}
			]
		}
	];

	if (process.platform === 'darwin') {
		const name = app.getName();
		template.unshift({
			label: name,
			submenu: [
				{
					role: 'about'
				}, {
					type: 'separator'
				}, {
					role: 'services',
					submenu: []
				}, {
					type: 'separator'
				}, {
					role: 'hide'
				}, {
					role: 'hideothers'
				}, {
					role: 'unhide'
				}, {
					type: 'separator'
				}, {
					role: 'quit'
				}
			]
		});
		// Edit menu.
		template[2].submenu.push(
			{
				type: 'separator'
			}, {
				label: 'Speech',
				submenu: [
					{
						role: 'startspeaking'
					}, {
						role: 'stopspeaking'
					}
				]
			}
		);
		// Window menu.
		template[4].submenu = [
			{
				label: 'Close Tab',
				accelerator: 'CmdOrCtrl+W',
				click: closeTabFunc
			}, {
				label: 'Minimize',
				accelerator: 'CmdOrCtrl+M',
				role: 'minimize'
			}, {
				label: 'Zoom',
				role: 'zoom'
			}, {
				type: 'separator'
			}, {
				label: 'Bring All to Front',
				role: 'front'
			}
		];
	}

	const menu = Menu.buildFromTemplate(template);
	Menu.setApplicationMenu(menu);
}

function openWindow() {
	console.log('openWindow');
	electron.session.defaultSession.clearCache().then(async function () {
		var curWindow = new BrowserWindow({
			x: mainWindowState.x
			, y: mainWindowState.y
			, width: mainWindowState.width
			, height: mainWindowState.height
			, webPreferences: {
				contextIsolation: true
				, sandbox: false
				, nodeIntegration: true
				, nodeIntegrationInSubFrames: true
				, enableRemoteModule: true
				, preload: path.join(__dirname, 'preload_main.js')
				, webviewTag: true
			}
		});
		curWindow.on('close', beforeWindowClose);
		mainWindows.push(curWindow);
		mainWindowState.manage(curWindow);

		// must close update window after main window is opened
		if (globalUpdateWindow) {
			globalUpdateWindow.close();
			globalUpdateWindow = null;
		}

		curWindow.loadURL('http://127.0.0.1:' + int_postage_port + '/postage/' +
			(process.argv.indexOf('--postage-test') > -1 ? 'test/' : '') + 'index.html' +
			(process.argv.indexOf('--postage-test') > -1 ? '?seq_numbers=true&_http_auth=true&http_file=true&http_upload=true&http_export=true&ws_raw=true&ws_tab=true&ws_select=true&ws_insert=true&ws_update=true&ws_delete=true' : ''),
			{ 'extraHeaders': 'pragma: no-cache\n' });

		curWindow.webContents.on('will-prevent-unload', function (event) {
			let choice = dialog.showMessageBoxSync(curWindow, {
				type: 'question',
				buttons: ['Leave', 'Stay'],
				title: 'Are you sure?',
				message: 'Changes you made may not be saved.',
				defaultId: 0,
				cancelId: 1
			});
			let leave = (choice === 0);
			if (leave) {
				event.preventDefault();
			}
		});

		// Emitted when the window is closed.
		curWindow.on('closed', function () {
			mainWindows.splice(mainWindows.indexOf(curWindow), 1);
		});

		curWindow.webContents.on('will-navigate', handleRedirect);
		curWindow.webContents.on('new-window', handleRedirect);
		curWindow.webContents.on('context-menu', function (event, props) {
			const template = [
				{
					role: 'undo'
				}, {
					role: 'redo'
				}, {
					type: 'separator'
				}, {
					role: 'cut',
					enabled: props.canCut
				}, {
					role: 'copy',
					enabled: props.canCopy
				}, {
					role: 'paste',
					enabled: props.canPaste
				}, {
					role: 'delete',
					enabled: props.canDelete
				}, {
					role: 'selectall',
					enabled: props.canSellectAll
				}
			];

			if (isDev) {
				template.push({
					type: 'separator'
				}, {
					label: 'Inspect Element',
					click: function (item, win) {
						win.webContents.inspectElement(props.x, props.y);
						if (win.webContents.isDevToolsOpened()) {
							win.webContents.devToolsWebContents.focus();
						}
					}
				});
			}

			const contextMenu = electron.Menu.buildFromTemplate(template);

			contextMenu.popup(curWindow);
		});

		setMenu();
	});
}

// This method will be called when Electron has finished
// initialization and is ready to create browser windows.
// Some APIs can only be used after this event occurs.
app.whenReady().then(async function () {
	electron.session.defaultSession = electron.session.fromPartition('persist:postage-session123', { cache: false });
	mainWindowState = windowStateKeeper({
		defaultWidth: 1024,
		defaultHeight: 768,
		path: os.homedir() + '/.postage/',
		file: 'main-window-state.json'
	});
	console.log('this is a test');

	if (bolPostageIsReady) {
		console.log('this is a test2');
		createUpdateWindow();
	}
});

// app.on('ready', appStart);

// Quit when all windows are closed.
app.on('window-all-closed', function () {
	// On OS X it is common for applications and their menu bar
	// to stay active until the user quits explicitly with Cmd + Q
	if (process.platform !== 'darwin') {
		app.quit();
	}
});

app.on('activate', function () {
	// On OS X it's common to re-create a window in the app when the
	// dock icon is clicked and there are no other windows open.
	if (mainWindows.length === 0 && !globalUpdateWindow) {
		openWindow();
	}
});

ipcMain.handle('openPath', function (event, message) {
	console.log(message.replace(/\{\{TABPATH\}\}/gi,
		process.platform === 'win32'
		? path.normalize(app.getPath('userData') + '\\sql\\')
		: path.normalize(app.getPath('home') + '/.postage/sql/')
	));
	electron.shell.openExternal(message.replace(/\{\{TABPATH\}\}/gi,
		process.platform === 'win32'
		? path.normalize(app.getPath('userData') + '\\sql\\')
		: path.normalize(app.getPath('home') + '/.postage/sql/')
	));
});

ipcMain.handle('updateReady', function (event) {
    if (isDev) {
        globalUpdateWindow.webContents.send('message', 'Development environment, starting...')

        setTimeout(function() {
            openWindow();
            //globalUpdateWindow.webContents.send('ready-to-share', getGUID())
        }, 1000);
    } else {
        autoUpdater.on('checking-for-update', function () {
            globalUpdateWindow.webContents.send('message', 'Checking for update...');
        })
        
        autoUpdater.on('update-available', function () {
            globalUpdateWindow.webContents.send('message', 'Update available');
        })
        
        autoUpdater.on('update-not-available', function () {
            globalUpdateWindow.webContents.send('message', 'No update, starting...');

            setTimeout(function() {
                openWindow();
                //globalUpdateWindow.webContents.send('ready-to-share', getGUID())
            }, 1000);
        });
        
        autoUpdater.on('error', function (strError) {
            console.error('There was a problem updating the application');
            console.error(strError);

            if ((/the command is disabled and cannot be executed/i).test(strError)) {
                globalUpdateWindow.webContents.send(
                    'message'
                    , 'A system prompt will appear.'
                    , 'Please enter administrator username and password, then click "Add Helper" to finish updating'
                );
            } else {
                globalUpdateWindow.webContents.send('message', strError);

				setTimeout(function() {
					openWindow();
					//globalUpdateWindow.webContents.send('ready-to-share', getGUID())
				}, 1000);
            }
        });
        
        autoUpdater.on('download-progress', function (progressObj) {
            globalUpdateWindow.webContents.send('message', 'Downloading...');
        });

        autoUpdater.on('update-downloaded', function (info) {
            globalUpdateWindow.webContents.send('message', 'Update downloaded, will restart shortly');

            //update automatically
            setTimeout(function() {
                autoUpdater.quitAndInstall();
            }, 1000);
        })

        autoUpdater.checkForUpdates();
    }
});

function createUpdateWindow() {
    //create the browser window
    globalUpdateWindow = new BrowserWindow({
        width: 300
        , height: 350
        //width: 800
        //, height: 1000
        , icon: 'logo.png'
        , webPreferences: {
            contextIsolation: true
            , sandbox: true
            , nodeIntegration: false
            , enableRemoteModule: false
            , preload: path.join(__dirname, 'preload_update.js')
        }
    })

    //load the index.html of the app
    globalUpdateWindow.loadURL('http://127.0.0.1:' + int_postage_port + '/postage/update/update.html')

    //remove window menu
    globalUpdateWindow.removeMenu();

    //open the devtools
    // globalUpdateWindow.webContents.openDevTools()

    //close all other windows
	mainWindows.forEach(function (curWindow) {
		curWindow.close();
	});
}
